/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import java.util.Stack;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ExtendedModifyEvent;
import org.eclipse.swt.custom.ExtendedModifyListener;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;

/**
 * Hack to fix the seemingly broken undo-redo functionality of SourceViewer.
 * 
 * @author aravantinos
 */
// TODO (AD): Please rename to SourceEditorUndoRedo (Easier to read).
public class SourceEditorUndoRedo implements KeyListener, ExtendedModifyListener {
	/**
	 * Encapsulation of the Undo and Redo stack(s).
	 */
	private static class UndoRedoStack<T> {

		/** Command stack in the "undo direction". */
		private Stack<T> undo;

		/** Command stack in the "redo direction". */
		private Stack<T> redo;

		/** Constructor. */
		public UndoRedoStack() {
			undo = new Stack<T>();
			redo = new Stack<T>();
		}

		/** Adds a command to the undo stack. */
		public void pushUndo(T delta) {
			undo.add(delta);
		}

		/** Adds a command to the redo stack. */
		public void pushRedo(T delta) {
			redo.add(delta);
		}

		/** Pops a command from the undo stack. */
		public T popUndo() {
			T res = undo.pop();
			return res;
		}

		/** Pops a command from the redo stack. */
		public T popRedo() {
			T res = redo.pop();
			return res;
		}

		/** Empties the redo stack. */
		public void clearRedo() {
			redo.clear();
		}

		/** Returns true if the undo stack is not empty. */
		public boolean hasUndo() {
			return !undo.isEmpty();
		}

		/** Returns true if the redo stack is not empty. */
		public boolean hasRedo() {
			return !redo.isEmpty();
		}
	}

	/** The editor GUI itself. */
	private StyledText editor;

	/** The local undo/redo command stack. */
	private UndoRedoStack<ExtendedModifyEvent> stack;

	/** Flag to know whether we are in the course of an undo event. */
	private boolean isUndo;

	/** Flag to know whether we are in the course of an redo event. */
	private boolean isRedo;

	/**
	 * Creates a new instance of this class. Automatically starts listening to
	 * corresponding key and modify events coming from the given
	 * <var>editor</var>.
	 * 
	 * @param editor
	 *            the text field to which the Undo-Redo functionality should be
	 *            added
	 */
	public SourceEditorUndoRedo(StyledText editor) {
		editor.addExtendedModifyListener(this);
		editor.addKeyListener(this);

		this.editor = editor;
		stack = new UndoRedoStack<ExtendedModifyEvent>();
	}

	/** {@inheritDoc} */
	@Override
	public void keyPressed(KeyEvent e) {
		// Listen to CTRL+Z for Undo, to CTRL+Y or CTRL+SHIFT+Z for Redo
		if((e.stateMask & SWT.MOD1) > 0 && !((e.stateMask & SWT.ALT) > 0)) {
			boolean isShift = (e.stateMask & SWT.SHIFT) > 0;
			if(!isShift && e.keyCode == 'z') {
				undo();
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void keyReleased(KeyEvent e) {
		// Listen to CTRL+Z for Undo, to CTRL+Y or CTRL+SHIFT+Z for Redo
		if((e.stateMask & SWT.MOD1) > 0 && !((e.stateMask & SWT.ALT) > 0)) {
			boolean isShift = (e.stateMask & SWT.SHIFT) > 0;
			if(!isShift && e.keyCode == 'z') {
				undo();
			} else if(!isShift && e.keyCode == 'y' || isShift && e.keyCode == 'z') {
				redo();
			}
		}
	}

	/**
	 * Creates a corresponding Undo or Redo step from the given event and pushes
	 * it to the stack. The Redo stack is, logically, emptied if the event comes
	 * from a normal user action.
	 * 
	 * @param event
	 * @see org.eclipse.swt.custom.ExtendedModifyListener#modifyText(org.eclipse.
	 *      swt.custom.ExtendedModifyEvent)
	 */
	@Override
	public void modifyText(ExtendedModifyEvent event) {
		if(isUndo) {
			stack.pushRedo(event);
		} else { // is Redo or a normal user action
			stack.pushUndo(event);
			if(!isRedo) {
				stack.clearRedo();
			}
		}
	}

	/**
	 * Performs the Undo action. A new corresponding Redo step is automatically
	 * pushed to the stack.
	 */
	private void undo() {
		if(stack.hasUndo()) {
			isUndo = true;
			revertEvent(stack.popUndo());
			isUndo = false;
		}
	}

	/**
	 * Performs the Redo action. A new corresponding Undo step is automatically
	 * pushed to the stack.
	 */
	private void redo() {
		if(stack.hasRedo()) {
			isRedo = true;
			revertEvent(stack.popRedo());
			isRedo = false;
		}
	}

	/**
	 * Reverts the given modify event, in the way as the Eclipse text editor
	 * does it.
	 * 
	 * @param event
	 */
	private void revertEvent(ExtendedModifyEvent event) {
		editor.replaceTextRange(event.start, event.length, event.replacedText);
		// (causes the modifyText() listener method to be called)

		editor.setSelectionRange(event.start, event.replacedText.length());
	}
}
